"""
Typer Configuration Utilities
"""

from typer import BadParameter, CallbackParam, Context

from .__typing import ConfigLoader, ConfigParameterCallback, TyperParameterValue
from .loaders import dotenv_loader, json_loader, toml_loader, yaml_loader


def conf_callback_factory(loader: ConfigLoader) -> ConfigParameterCallback:
    """Typer configuration callback factory.

    Parameters
    ----------
    loader : ConfigLoader
        Config loader function that takes the value passed to the typer CLI and
        returns a dictionary that is applied to the click context's default map.

    Returns
    -------
    ConfigParameterCallback
        Configuration parameter callback function.
    """

    def _callback(
        ctx: Context, param: CallbackParam, param_value: TyperParameterValue
    ) -> TyperParameterValue:
        """Generated typer config parameter callback.

        Parameters
        ----------
        ctx : typer.Context
            typer context (automatically passed)
        param : typer.CallbackParam
            typer callback parameter (automatically passed)
        param_value : TyperParameterValue
            parameter value passed to typer (automatically passed)

        Returns
        -------
        TyperParameterValue
            must return back the given parameter

        Raises
        ------
        typer.BadParameter
            bad parameter value
        """
        try:
            conf = loader(param_value)  # Load config file
            ctx.default_map = ctx.default_map or {}  # Initialize the default map
            ctx.default_map.update(conf)  # Merge the config Dict into default_map
        except Exception as ex:
            raise BadParameter(str(ex), ctx=ctx, param=param) from ex
        return param_value

    return _callback


yaml_conf_callback: ConfigParameterCallback = conf_callback_factory(yaml_loader)
"""YAML configuration callback for a typer parameter.

Parameters
----------
ctx : typer.Context
    typer context (automatically passed)
param : typer.CallbackParam
    typer callback parameter (automatically passed)
param_value : TyperParameterValue
    parameter value passed to typer (automatically passed)

Returns
-------
TyperParameterValue
    must return back the given parameter

Raises
------
typer.BadParameter
    bad parameter value
"""

json_conf_callback: ConfigParameterCallback = conf_callback_factory(json_loader)
"""JSON configuration callback for a typer parameter.

Parameters
----------
ctx : typer.Context
    typer context (automatically passed)
param : typer.CallbackParam
    typer callback parameter (automatically passed)
param_value : TyperParameterValue
    parameter value passed to typer (automatically passed)

Returns
-------
TyperParameterValue
    must return back the given parameter

Raises
------
typer.BadParameter
    bad parameter value
"""


toml_conf_callback: ConfigParameterCallback = conf_callback_factory(toml_loader)
"""TOML configuration callback for a typer parameter.

Parameters
----------
ctx : typer.Context
    typer context (automatically passed)
param : typer.CallbackParam
    typer callback parameter (automatically passed)
param_value : TyperParameterValue
    parameter value passed to typer (automatically passed)

Returns
-------
TyperParameterValue
    must return back the given parameter

Raises
------
typer.BadParameter
    bad parameter value
"""

dotenv_conf_callback: ConfigParameterCallback = conf_callback_factory(dotenv_loader)
"""Dotenv configuration callback for a typer parameter.

Parameters
----------
ctx : typer.Context
    typer context (automatically passed)
param : typer.CallbackParam
    typer callback parameter (automatically passed)
param_value : TyperParameterValue
    parameter value passed to typer (automatically passed)

Returns
-------
TyperParameterValue
    must return back the given parameter

Raises
------
typer.BadParameter
    bad parameter value
"""



"""
Configuration File Loaders.

These loaders must implement the `typer_config.__typing.ConfigLoader` interface.
"""
import json
import sys

from .__typing import (
    ConfigDict,
    ConfigDictAccessorPath,
    ConfigLoader,
    NoArgCallable,
    TyperParameterValue,
)

USING_TOMLLIB = False
TOML_MISSING = True
YAML_MISSING = True
DOTENV_MISSING = True


if sys.version_info >= (3, 11):  # pragma: no cover
    import tomllib  # type: ignore

    TOML_MISSING = False
    USING_TOMLLIB = True
else:  # pragma: no cover
    try:
        # Third-party toml parsing library
        import toml

        TOML_MISSING = False

    except ImportError:
        pass


try:  # pragma: no cover
    import yaml

    YAML_MISSING = False
except ImportError:  # pragma: no cover
    pass

try:  # pragma: no cover
    import dotenv

    DOTENV_MISSING = False
except ImportError:  # pragma: no cover
    pass


def subpath_loader(
    loader: ConfigLoader, dictpath: ConfigDictAccessorPath
) -> ConfigLoader:
    """Modify a loader to return a subpath of the dictionary from file.

    Parameters
    ----------
    loader : ConfigLoader
        loader to modify
    dictpath : ConfigDictAccessorPath
        path to the section of the dictionary to return

    Returns
    -------
    ConfigLoader
        sub dictionary loader
    """

    def _loader(param_value: str) -> ConfigDict:
        # get original ConfigDict
        conf: ConfigDict = loader(param_value)

        # get subpath of dictionary
        for path in dictpath:
            conf = conf.get(path, {})
        return conf

    return _loader


def default_value_loader(
    loader: ConfigLoader, value_getter: NoArgCallable
) -> ConfigLoader:
    """Modify a loader to use a default value if the passed value is false-ish

    Parameters
    ----------
    loader : ConfigLoader
        loader to modify
    value_getter : NoArgCallable
        function that returns default value

    Returns
    -------
    ConfigLoader
        modified loader
    """

    def _loader(param_value: str) -> ConfigDict:
        # parameter value was not specified by user
        if not param_value:
            param_value = value_getter()

        conf: ConfigDict = loader(param_value)

        return conf

    return _loader


def yaml_loader(param_value: TyperParameterValue) -> ConfigDict:
    """YAML file loader

    Parameters
    ----------
    param_value : TyperParameterValue
        path of YAML file

    Returns
    -------
    ConfigDict
        dictionary loaded from file
    """

    if YAML_MISSING:  # pragma: no cover
        raise ModuleNotFoundError("Please install the pyyaml library.")

    with open(param_value, "r", encoding="utf-8") as _file:
        conf: ConfigDict = yaml.safe_load(_file)

    return conf


def json_loader(param_value: TyperParameterValue) -> ConfigDict:
    """JSON file loader

    Parameters
    ----------
    param_value : TyperParameterValue
        path of JSON file

    Returns
    -------
    ConfigDict
        dictionary loaded from file
    """

    with open(param_value, "r", encoding="utf-8") as _file:
        conf: ConfigDict = json.load(_file)

    return conf


def toml_loader(param_value: TyperParameterValue) -> ConfigDict:
    """TOML file loader

    Parameters
    ----------
    param_value : TyperParameterValue
        path of TOML file

    Returns
    -------
    ConfigDict
        dictionary loaded from file
    """

    if TOML_MISSING:  # pragma: no cover
        raise ModuleNotFoundError("Please install the toml library.")

    conf: ConfigDict = {}

    if USING_TOMLLIB:  # pragma: no cover
        with open(param_value, "rb") as _file:
            conf = tomllib.load(_file)  # type: ignore
    else:  # pragma: no cover
        with open(param_value, "r", encoding="utf-8") as _file:
            conf = toml.load(_file)  # type: ignore

    return conf


def dotenv_loader(param_value: TyperParameterValue) -> ConfigDict:
    """Dotenv file loader

    Parameters
    ----------
    param_value : TyperParameterValue
        path of YAML file

    Returns
    -------
    ConfigDict
        dictionary loaded from file
    """

    if DOTENV_MISSING:  # pragma: no cover
        raise ModuleNotFoundError("Please install the python-dotenv library.")

    with open(param_value, "r", encoding="utf-8") as _file:
        # NOTE: I'm using a stream here so that the loader
        # will raise an exception when the file doesn't exist.
        conf: ConfigDict = dotenv.dotenv_values(stream=_file)

    return conf



"""
Data and Function types.
"""

import sys
from typing import Any, Callable, Dict, Iterable

from typer import CallbackParam, Context

# Handle some imports based on python version
if sys.version_info < (3, 10):  # pragma: no cover
    from typing_extensions import TypeAlias
else:  # pragma: no cover
    from typing import TypeAlias

# Data types
TyperParameterName: TypeAlias = str
"""Typer CLI parameter name."""

TyperParameterValue: TypeAlias = Any
"""Typer CLI parameter value."""

ConfigDict: TypeAlias = Dict[TyperParameterName, Any]
"""Configuration dictionary to be applied to the click context default map."""

ConfigDictAccessorPath: TypeAlias = Iterable[str]
"""Configuration dictionary accessor path."""

# Function types
ConfigLoader: TypeAlias = Callable[[TyperParameterValue], ConfigDict]
"""Configuration loader function."""

ConfigParameterCallback: TypeAlias = Callable[
    [Context, CallbackParam, TyperParameterValue], TyperParameterValue
]
"""Typer config parameter callback function."""

NoArgCallable: TypeAlias = Callable[[], Any]
"""No argument callable."""
